from past.builtins import cmp
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
import logging
from nomadcore import compact_sha
import json
import os, re
from nomadcore.json_support import jsonCompactS, jsonCompactD, jsonIndentD
from io import open
# We removed the old nomad-meta-info project and the metainfo module respectively. Most of the
# functionality in this module wont work anymore.
# import metainfo
"""objects to handle a local InfoKinds with unique name (think self written json)"""

class InfoKindEl(object):
    """Info kind (tipically from a file, without shas but with locally unique names)"""
    __slots__ = ["name","description","kindStr","units","superNames","dtypeStr", "repeats", "shape", "extra_args", "package"]
    IGNORE_EXTRA_ARGS = 1
    ADD_EXTRA_ARGS = 2
    RAISE_IF_EXTRA_ARGS = 3

    def __init__(self, name, description, kindStr = "type_document_content", units = None, superNames = None,
            dtypeStr = None, shape = None, package = None, extraArgsHandling = ADD_EXTRA_ARGS, repeats = None, **extra_args):
        if superNames is None:
            superNames = []
        self.name = name
        self.description = description
        self.kindStr = kindStr
        self.superNames = superNames
        self.units = units
        self.dtypeStr = dtypeStr
        self.package = package
        if dtypeStr in ["None", "null"]:
            self.dtypeStr = None
        self.shape = shape
        self.repeats = repeats
        if extraArgsHandling == self.ADD_EXTRA_ARGS:
            self.extra_args = extra_args
        elif extraArgsHandling == self.IGNORE_EXTRA_ARGS:
            self.extra_args = {}
        else:
            raise Exception("extra arguments to InfoKindEl:" + str(extra_args))

    def __eq__(o1, o2):
        try:
            if not (o1.name == o2.name and o1.description == o2.description and o1.kindStr == o2.kindStr and
                o1.units == o2.units and o1.shape == o2.shape):
                return False
            if o1.dtypeStr != o2.dtypeStr:
                return False
            if o1.repeats != o2.repeats:
                return False
            if o1.extra_args != o2.extra_args:
                return False
            if o1.superNames == o2.superNames:
                return True
            if len(o1.superNames) != len(o2.superNames):
                return False
            if o1.superNames[0] != o2.superNames[0]:
                return False
            a1 = o1.superNames[1:]
            a2 = o2.superNames[1:]
            a1.sort()
            a2.sort()
            for i in range(len(a1)):
                if a1[i] != a2[i]:
                    return False
            return True
        except:
            raise
            return False

    def __cmp__(k1, k2):
        c = cmp(k1.name, k2.name)
        if c != 0: return c
        c = cmp(k1.kindStr, k2.kindStr)
        if c != 0: return c
        c = cmp(k1.description, k2.description)
        if c != 0: return c
        if len(k1.superNames) > 0:
            if len(k2.superNames) > 0:
                c = cmp(k1.superNames[0], k2.superNames[0])
                if c != 0: return c
                s1 = k1.superNames[1:]
                s2 = k2.superNames[1:]
                c = cmp(s1, s2)
                if c != 0: return c
            else:
                return 1
        elif len(k2.superNames) > 0:
            return -1
        if c != 0: return c
        c = cmp(k1.units, k2.units)
        if c != 0: return c
        c = cmp(k1.dtypeStr, k2.dtypeStr)
        if c != 0: return c
        c = cmp(k1.repeats, k2.repeats)
        if c != 0: return c
        c = cmp(k1.shape, k2.shape)
        if c != 0: return c
        if k1.extra_args == k2.extra_args:
            return 0
        if k1.extra_args is None:
            return 1
        if k2.extra_args is None:
            return -1
        extraK1 = list(k1.extra_args.keys())
        extraK1.sort()
        extraK2 = list(k2.extra_args.keys())
        extraK2.sort()
        i = 0
        while (i < len(extraK1) and i < len(extraK2)):
            kk1 = extraK1[i]
            kk2 = extraK2[i]
            c = cmp(kk1, kk2)
            if c != 0: return c # use -c ?
            c = cmp(k1.extra_args[kk1], k2.extra_args[kk2])
            if c != 0: return c
        c = cmp(len(extraK1), len(extraK2))
        return c

    def __ne__(o1, o2):
        return not o1.__eq__(o2)

    def prepare(self, env):
        if len(self.superNames) > 1:
            a = self.superNames[1:]
            a.sort(lambda x, y: cmp(env.gidOf(x, precalculatedGid = True), env.gidOf(y, precalculatedGid = True)))
            self.superNames[1:] = a

    def evalGid(self, env):
        self.prepare(env)
        sha = env.newSha()
        self.serialize(env,sha.update, precalculatedGid = True, selfGid = False)
        return 'p' + sha.b64digest()[:28]

    def serialize(self, env, writeOut, subGids = True, addExtraArgs = True, precalculatedGid = False, selfGid = True):
        d = self.toDict(env, subGids = subGids, addExtraArgs = addExtraArgs, precalculatedGid = precalculatedGid, selfGid = selfGid)
        jsonCompactD(d, writeOut)

    def toDict(self, env = None, addExtraArgs = True, inlineExtraArgs = True , selfGid = False, subGids = False, precalculatedGid = False):
        res = {
            "description": self.description,
            "name": self.name,
            "superNames": self.superNames,
        }
        try:
            if self.kindStr != "type_document_content":
                if self.kindStr is None or self.kindStr == "":
                    res["kindStr"] = "MetaType"
                else:
                    res["kindStr"] = self.kindStr
            if env:
                if selfGid:
                    res["gid"] = env.gidOf(self.name, precalculatedGid = precalculatedGid)
                if subGids:
                    res["superGids"] = [ env.gidOf(sName, precalculatedGid = precalculatedGid) for sName in self.superNames ]
            elif subGids or selfGid:
                raise Exception("env required in toDict for subGids or selfGid")
            if self.units is not None:
                res["units"] = self.units
            if self.dtypeStr is not None:
                res["dtypeStr"] = self.dtypeStr
            if self.repeats is not None:
                res["repeats"] = self.repeats
            if self.shape is not None:
                res["shape"] = self.shape
            if addExtraArgs:
                if inlineExtraArgs:
                    res.update(self.extra_args)
                else:
                    res["extraArgs"] = self.extra_args
        except:
            logging.exception("error in InfoKindEl.toDict, partial dict is %s", res)
        return res

    def __unicode__(self):
        s = StringIO.StringIO()
        self.serialize(s)
        return s.string

class RelativeDependencySolver(object):
    def __init__(self):
        self.deps = {}

    def __call__(self, infoKindEnv, source, dep):
        if "metainfoPath" in dep:
            basePath = metainfo.__file__
            path = dep["metainfoPath"]
        elif "relativePath" in dep:
            basePath = source.get('path')
            path = dep["relativePath"]
        else:
            raise Exception('Invalid dependency for relativeDependencySolver there must be a relativePath or metainfoPath')

        if basePath:
            baseDir = os.path.dirname(os.path.abspath(basePath))
        else:
            baseDir = os.getcwd()
        dPath = os.path.realpath(os.path.join(baseDir, path))
        if dPath in self.deps:
            return self.deps[dPath]
        depInfo = None
        depIKEnv = InfoKindEnv(path = dPath, dependencyLoader=infoKindEnv.dependencyLoader)
        self.deps[dPath] = depIKEnv
        with open(dPath, encoding="utf-8") as f:
            try:
                depInfo = json.load(f)
            except:
                logging.exception("Error while loading dependency %s" % f)
                raise
        if depInfo:
            depIKEnv.fromJsonList(depInfo, name = os.path.basename(dPath), source = { 'path': dPath }, dependencyLoad = False)
        return depIKEnv


class InfoKindEnv(object):
    """An environment keeping locally unique InfoKinds and their gids"""
    def __init__(self, infoKinds = None, name = None, description = None, newSha = compact_sha.sha512, gids = None,
            dependencyLoader = None, path = None, uri = None, deps = None):
        self.newSha = newSha
        self.clear()
        self.name = name
        self.description = description
        self.dependencyLoader = dependencyLoader
        if dependencyLoader is None:
            self.dependencyLoader = RelativeDependencySolver()
        self.path = path
        self.uri  = uri
        if not infoKinds is None:
            for ik in infoKinds:
                self.addInfoKindEl(ik)
        if not gids is None:
            self.gids = gids
        if deps:
            self.deps = deps

    def __str__(self):
        if self.path:
            return "InfoKindEnv loaded from {}".format(self.path)

    def clear(self):
        self.gids = {}
        self.infoKinds = {}
        self.deps = []

    def depNames(self):
        res = set()
        for dep in self.deps:
            for name in dep.infoKinds.keys():
                res.add(name)
        return res

    def noDepNames(self):
        return set(self.infoKinds.keys()).difference(self.depNames())

    def embedDeps(self):
        hidden = []
        duplicate = set()
        for dep in self.deps:
            for name, ikEl in dep.infoKinds.items():
                oldVal=self.infoKinds.get(name, None)
                if oldVal is None:
                    self.infoKinds[name] = ikEl
                elif ikEl != oldVal:
                    hidden.append(ikEl)
                else:
                    duplicate.add(name)
        return { "hidden": hidden, "duplicate": duplicate }

    def addInfoKindEl(self, infoKind):
        if infoKind.name in self.infoKinds and infoKind != self.infoKinds[infoKind.name]:
            raise Exception('InfoKindEnv has collision for name {0}: {1} vs {2}'
                    .format(infoKind.name, infoKind.package, self.infoKinds[infoKind.name].package))
        self.infoKinds[infoKind.name] = infoKind

    def addDependenciesFrom(self, infoKindEnv):
        toAdd = set(self.infoKinds.keys())
        missing = set()
        while len(toAdd):
            ikName = toAdd.pop()
            ik = self.infoKinds.get(ikName,None)
            if ik is None:
                depInfoKindEl = infoKindEnv.infoKinds.get(ikName, None)
                if depInfoKindEl:
                    self.infoKinds[ikName] = depInfoKindEl
                    toAdd.add(ikName)
                else:
                    missing.add(ikName)
            else:
                for dep in ik.superNames:
                    if not dep in self.infoKinds:
                        toAdd.add(dep)
        return missing

    def gidOf(self, name, precalculatedGid=False):
        res = self.gids.get(name,None)
        if res is None:
            if precalculatedGid:
                raise Exception("non precalculated gid for %s" % name)
            res = self.calcGid(name)
        return res

    def calcGid(self, name):
        inProgress = []
        toDo = [name]
        hasPending = False
        for i in range(2):
            while len(toDo) > 0:
                if not hasPending and inProgress:
                    now = inProgress.pop()
                else:
                    now = toDo.pop()
                if now in self.gids and now in inProgress:
                    inProgress.remove(now)
                hasPending = False
                nowVal = self.infoKinds.get(now, None)
                if nowVal is None:
                    raise Exception("while calculating gid of %r found unknown key %r" % (name, now))
                for subName in nowVal.superNames:
                    if subName in self.gids:
                        continue
                    hasPending = True
                    if subName in toDo:
                        toDo.remove(subName)
                    if subName in inProgress:
                        raise Exception('found loop to %s evaluating %s, currently in progress: %s' % (subName, now, inProgress))
                    toDo.append(subName)
                if not hasPending:
                    self.gids[now] = nowVal.evalGid(self)
                    if now in inProgress:
                        inProgress.remove(now)
                else:
                    if now in inProgress:
                        raise Exception('found loop to %s, currently in progress: %s' % (now, inProgress))
                    inProgress.append(now)
            toDo = list(inProgress)
        return self.gids[name]

    def keyDependsOnKey(self, k1Name, k2Name):
        """partial ordering given by the dependencies
        1: k1Name depends on k2Name
        0: k1Name == k2Name
        -1: k2Name depends on k1Name
        None: no dependency"""
        if k1Name == k2Name: return 0
        k1 = self.infoKinds[k1Name]
        k2 = self.infoKinds[k2Name]
        if k1.superNames != k2.superNames:
            allSuperK1 = set()
            toDoK1 = list(k1.superNames)
            allSuperK2 = set()
            toDoK2 = list(k2.superNames)
            while (len(toDoK1) > 0 or len(toDoK2) > 0):
                if len(toDoK1) > 0:
                    el1Name = toDoK1.pop()
                    if k2Name == el1Name:
                        return 1
                    el1 = self.infoKinds[el1Name]
                    if el1.kindStr in self and not el1.kindStr in allSuperK1:
                        toDoK1.append(el1.kindStr)
                    for subEl in el1.superNames:
                        if not subEl in allSuperK1:
                            toDoK1.append(subEl)
                    allSuperK1.update(el1.superNames)
                if len(toDoK2) > 0:
                    el2Name = toDoK2.pop()
                    if k1Name == el2Name:
                        return -1
                    el2 = self.infoKinds[el2Name]
                    if el2.kindStr in self and not el2.kindStr in allSuperK2:
                        toDoK2.append(el2.kindStr)
                    for subEl in el2.superNames:
                        if not subEl in allSuperK2:
                            toDoK2.append(subEl)
                    allSuperK2.update(el2.superNames)
        return None

    def __contains__(self, name):
        "if an item with the given name is contained in this environment"
        return name in self.infoKinds

    def __len__(self):
        """returns the number of InfoKindEl stored in this environment"""
        return len(self.infoKinds)

    def __getitem__(self, name):
        """returns a dictionary representing the entry with the given name, or None if it does not exist"""
        ikEl = self.infoKinds.get(name, None)
        if ikEl:
            return ikEl.toDict(self)
        return None

    def infoKindEls(self):
        return list(self.infoKinds.values())

    def infoKindEl(self, name):
        """returns the InfoKindEl with the given name, or None if it does not exist"""
        return self.infoKinds.get(name, None)

    def calcGids(self):
        for k in self.infoKinds.keys():
            if not k in self.gids:
                self.gids[k]=self.calcGid(k)

    def serialize(self, writeOut, subGids = True, selfGid = True):
        infoKinds = self.sortedIKs()
        writeOut("""{
  "type": "nomad_meta_info_1_0",
  "description": """)
        if self.description:
            jsonIndentD(self.description, writeOut, extraIndent = 4)
        else:
            writeOut('""')
        writeOut(',\n')
        if not self.path:
            baseDir = os.getcwd()
        else:
            baseDir = os.path.normpath(os.path.dirname(os.path.abspath(self.path)))
        depKeys = set()
        if self.deps:
            writeOut('  "dependencies": [ ')
            depColon = False
            for d in self.deps:
                path = d.path
                uri = d.uri
                depKeys.update(d.infoKinds.keys())
                if path:
                    path = os.path.normpath(os.path.abspath(path))
                    if path.startswith(baseDir) or not uri:
                        if depColon:
                            writeOut(", ")
                        else:
                            depColon = True
                        jsonIndentD({"relativePath": os.path.relpath(path, baseDir)}, writeOut, extraIndent = 4)
                        continue
                if uri:
                    if depColon:
                        writeOut(", ")
                    else:
                        depColon = True
                    jsonIndentD({"uri": uri}, writeOut, extraIndent = 4)
                    continue
                raise Exception("Dependency on serializable %s" % d)
            writeOut('],\n')
        addColon = False
        writeOut('  "metaInfos": [ ')
        for ik in infoKinds:
            if ik.name in depKeys:
                continue
            if addColon:
                writeOut(", ")
            else:
                addColon = True
            jsonIndentD(ik.toDict(env = self, subGids = subGids, selfGid = selfGid), writeOut, extraIndent = 4, check_circular = True)
        writeOut("]\n}\n")

    def sortedIKs(self):
        infoKinds = list(self.infoKinds.values())
        infoKinds.sort(lambda x, y: cmp(x.name.lower()+x.name, y.name.lower()+y.name))
        return infoKinds # self.sortAndComplete(infoKinds, ignoreMissing = True)


    def toJsonList(self, withGids):
        infoKinds = list(self.infoKinds.keys())
        # infoKinds.sort(lambda x, y: self.compareKeys(x.name, y.name))
        return [self.infoKinds[x].toDict(self,
                self if withGids else None) for x in infoKinds]

    def verifyGids(self, preserveAbsent=False):
        changes = {}
        oldGids = self.gids
        self.gids = {}
        self.calcGids()
        for k,v in oldGids.items():
            newVal = self.gids.get(k, None)
            if newVal is None:
                if preserveAbsent:
                    self.gids[k] = v
                else:
                    changes[k] = (v, None)
            elif v != newVal:
                changes[k] = (v, newVal)
        return changes

    def fromJsonList(self, jsonDict, name,  source, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, dependencyLoad=False):
        typeStr = jsonDict.get("type","nomad_meta_info_1_0")
        typeRe = re.compile(r"nomad_meta_info_(?P<major>[0-9]+)_(?P<minor>[0-9]+)$")
        self.name = name
        m = typeRe.match(typeStr)
        if not m:
            raise Exception("unexpected type '%s', expected nomad_meta_info_1_0" % typeStr)
        if int(m.group("major")) != 1:
            raise Exception("Unsupported major version %s, expeced 1")
        dependencies = jsonDict.get("dependencies",[])
        jsonList = jsonDict.get("metaInfos",[])
        self.description = jsonDict.get("description","")
        overwritten = []
        gidToCheck = {}
        deps = []
        for d in dependencies:
            if self.dependencyLoader is None:
                raise Exception("no dependencyLoader while loading local_in")
            dep = self.dependencyLoader(self, source, d)
            if dep:
                self.deps.append(dep)
        index = -1
        for ii in jsonList:
            index += 1
            val = dict(ii)
            if not "name" in ii:
                raise Exception("InfoKind at %d is without name: %s" % (index, ii) )
            oldVal=self.infoKinds.get(ii['name'],None)
            gid=None
            if "gid" in ii:
                gid = ii['gid']
                del val['gid']
            if "superGids" in ii:
                if not "superNames" in ii:
                    raise Exception("superGids without superNames in fromJsonList")
                superNames = ii["superNames"]
                superGids = ii["superGids"]
                if len(superNames) != len(superGids):
                    raise Exception("superGids incompatible with superNames in fromJsonList: %s vs %s" % (ii["superGids"], ii["superNames"]))
                toCheck = {}
                for i in range(len(superNames)):
                    assert not superNames[i] in toCheck.keys(), "duplicate superName %r in %r" % (superNames[i], ii["name"])
                    toCheck[superNames[i]] = superGids[i]
                gidToCheck[ii["name"]] = toCheck
                del val['superGids']
            val['extraArgsHandling'] = extraArgsHandling
            val['package'] = self.name
            ikEl = InfoKindEl(**val)
            if not oldVal is None and ikEl != oldVal:
                overwritten.append((oldVal, ikEl))
            if gid:
                self.gids[ii['name']] = gid
            self.infoKinds[ikEl.name] = ikEl
        res = { "overwritten": overwritten }
        if not dependencyLoad:
            res.update(self.embedDeps())
            for childName, gids in gidToCheck.items():
                for name, gid in gids.items():
                    if self.gidOf(name) != gid:
                        raise Exception("incompatible superGid for superName %s of %s (%s vs %s)" % (name, ii["name"], gid, self.gidOf(name)))
        if res.get("overwritten", False) or res.get("duplicate", False) or res.get("hidden", False):
            res["hasWarnings"] = True
        else:
            res["hasWarnings"] = res.get("hasWarnings", False)
        return res

    def sortAndComplete(self, propsToSort, ignoreMissing = False):
        """builds a list of properties in propsToSort, so that all the dependecies of each
        property are present before them"""
        toDo = list(propsToSort)
        done = set()
        deps = []
        res = []
        while len(toDo)>0:
            pAtt = toDo.pop()
            nameAtt = pAtt.name
            if nameAtt in done:
                continue
            deps = [nameAtt]
            while len(deps)>0:
                nameAtt = deps[-1]
                pAtt = self.infoKinds.get(nameAtt, None)
                if pAtt is None:
                    if ignoreMissing:
                        deps.pop()
                        done.add(nameAtt)
                        continue
                    else:
                        raise Exception("missing dependent InfoKindEl {0} following chain {1}".format(nameAtt, pAtt))
                hasDepsToDo = False
                kindStr = pAtt.kindStr
                kindType = self.infoKindEl(kindStr)
                for superName in pAtt.superNames:
                    if not superName in done:
                        if superName in deps:
                            raise Exception("circular dependency {0}, {1}".format(deps,superName))
                        deps.append(superName)
                        hasDepsToDo = True
                if kindType and not kindStr in done:
                    if kindStr in deps:
                        raise Exception("circular dependency in kindStr {0}, {1}".format(deps,kindStr))
                    deps.append(kindStr)
                    hasDepsToDo = True
                if not hasDepsToDo:
                    deps.pop()
                    res.append(pAtt)
                    done.add(nameAtt)
        return res

    def metaInfoNameWithAllSuper(self, name):
        """returns the meta info names of name and all its dependencies"""
        toAdd = set([name])
        res = set([name])
        while toAdd:
            e = toAdd.pop()
            for superName in self.infoKinds[e].superNames:
                if not superName in res:
                    res.add(superName)
                    toAdd.add(superName)
        return res

    def firstAncestorsByType(self, name):
        """Returns the first acestors of each type separated in roots and children.
        (scala conversion, could be improved a bit)"""
        metaInfoNames = self.metaInfoNameWithAllSuper(name)
        metaInfoNames.remove(name)
        mInfo = list(metaInfoNames)
        edges = {}
        for i, metaName in enumerate(mInfo):
            metaInfo = self.infoKinds[metaName]
            edges[i] = [mInfo.index(x) for x in metaInfo.superNames]
        typeGroups = {}
        for mIdx, metaName in enumerate(mInfo):
            kindStr = self.infoKinds[metaName].kindStr
            tNow = typeGroups.get(kindStr, None)
            if tNow is None:
                typeGroups[kindStr] = [mIdx]
            else:
                tNow.append(mIdx)
        childsByType = {}
        toDo  = set(range(len(mInfo)))
        while (toDo):
            now = toDo.pop()
            kindNow = self.infoKinds[mInfo[now]].kindStr
            toDo2 = set(edges[now])
            known2 = set(edges[now])
            while (toDo2):
                now2 = toDo2.pop()
                if (self.infoKinds[mInfo[now2]].kindStr == kindNow):
                    childs = childsByType.get(kindNow, None)
                    if childs:
                        childs.add(now2)
                    else:
                        childsByType[kindNow] = set([now2])
                    if now2 in toDo:
                        toDo.remove(now2)
                for el in edges[now2]:
                    if not el in known2:
                        toDo2.add(el)
                        known2.add(el)
        res = {}
        for typeName, allChilds in typeGroups.items():
            childs = childsByType.get(typeName, set())
            allForKind = set(allChilds)
            rootNames = [mInfo[x] for x in (allForKind - childs)]
            childNames = [mInfo[x] for x in childs]
            res[typeName] = (rootNames, childNames)
        return res

def loadJsonFile(filePath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None):
    env = InfoKindEnv(dependencyLoader = dependencyLoader, path = filePath, uri = uri)
    try:
        with open(filePath, encoding="utf-8") as f:
            o = json.load(f)
            warnings = env.fromJsonList(o, name = os.path.basename(filePath), source = {'path': filePath}, extraArgsHandling = extraArgsHandling)
    except:
        logging.exception("Error while loading file %s" % filePath)
        raise
    return env, warnings


def load_metainfo(filename, dependencyLoader=None, extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS, uri=None):
    """Loads a metainfo environment for a filename. The filename should not
    contain the full path, as the full path is resolved here and not by the
    caller.

    Args:
        filename: filename as a string.

    Returns:
        Tuple containing the metainfo environment, and any possible warnings
        that were encountered in the loading.
    """
    if os.path.isfile(filename):
        path = filename
    else:
        path = os.path.join(os.path.dirname(metainfo.__file__), filename)
    return loadJsonFile(path, dependencyLoader, extraArgsHandling, uri)

def loadJsonStream(fileStream, name = None, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, filePath = None, uri = None):
    if filePath is None:
        try:
            filePath = fileStream.name
        except:
            filePath = None
    if name is None and not filePath is None:
        name = os.path.basename(filePath)
    env = InfoKindEnv(dependencyLoader = dependencyLoader, name = name, path = filePath, uri = uri)
    try:
        o = json.load(fileStream)
        warnings = env.fromJsonList(o, source = {'path': filePath}, extraArgsHandling = extraArgsHandling)
    except:
        logging.exception("Error while loading file %s" % filePath)
        raise
    return env, warnings



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from functools import total_ordering
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT

logger = logging.getLogger("nomad.turbomoleParser")


@total_ordering
class _Forces(object):

    def __init__(self, elem, index):
        self.x = float("nan")
        self.y = float("nan")
        self.z = float("nan")
        self.elem = elem.capitalize()
        self.index = int(index)

    def __eq__(self, other):
        if isinstance(other, _Forces):
            return self.elem == other.elem and self.index == other.index
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, _Forces):
            if self.index < other.index:
                return True
            elif self.elem < other.elem:
                return True
        return False


class GradientParser(object):

    __re_atoms = re.compile(r"\s*([0-9]+)\s+([A-z]+)")
    __re_force = re.compile(r"\s*("+RE_FLOAT+")")

    def __init__(self, context, key="gradient"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.__forces = list()
        self.__forces_staging = list()

    def purge_data(self):
        self.__forces = list()
        self.__forces_staging = list()

    def set_backend(self, backend):
        self.__backend = backend

    # matcher generation methods

    # improve: check if Turbomole allows constraints in force calculations
    def write_forces(self, index_config):
        num_atoms = len(self.__forces)
        if num_atoms > 0:
            forces = np.ndarray(shape=(num_atoms, 3), dtype=float)
            forces[:, :] = 0.0
            for i, atom in enumerate(sorted(self.__forces)):
                forces[i, 0:3] = atom.x, atom.y, atom.z
            self.__backend.addArrayValues("atom_forces_raw", forces, gIndex=index_config,
                                          unit="hartree / bohr")

    def build_gradient_matcher(self):

        def extract_atoms(backend, groups):
            self.__forces_staging = list()
            match = self.__re_atoms.match(groups[0], 0)
            while match:
                index = int(match.group(1))
                element = match.group(0)
                to_add = _Forces(element, index)
                if to_add in self.__forces:
                    to_update = self.__forces[self.__forces.index(to_add)]
                    self.__forces_staging.append(to_update)
                else:
                    self.__forces_staging.append(to_add)
                    self.__forces.append(to_add)
                match = self.__re_atoms.match(groups[0], match.endpos)

        def extract_derivative(backend, groups):
            match = self.__re_force.match(groups[1], 0)
            axis = groups[0]
            index = 0
            while match:
                gradient_component = float(match.group(1).replace("D", "E").replace("d", "e"))
                setattr(self.__forces_staging[index], axis, gradient_component)
                match = self.__re_force.match(groups[1], match.endpos)
                index += 1

        def gradients():
            def derivative():
                return SM(r"dE/d([xyz])((?:\s+"+RE_FLOAT+")+)\s*$",
                          name="gradients",
                          startReAction=extract_derivative,
                          required=True
                          )
            return SM(r"\s*ATOM((?:\s+[0-9]+\s+[A-z]+)+)\s*$",
                      name="atoms list",
                      required=True,
                      repeats=True,
                      startReAction=extract_atoms,
                      subMatchers=[
                          derivative(),
                          derivative(),
                          derivative()
                      ]
                      )

        return SM(r"\s*cartesian\s+gradient\s+of\s+the\s+energy\s+\(hartree/bohr\)\s*$",
                  name="gradients header",
                  subMatchers=[
                      gradients()
                  ]
                  )

    def build_gradient_matcher_statpt(self):

        def extract_force(backend, groups):
            new_force = _Forces(groups[1], groups[0])
            new_force.x = float(groups[2])
            new_force.y = float(groups[3])
            new_force.z = float(groups[4])
            self.__forces.append(new_force)

        gradient = SM(r"\s*([0-9]+)\s+([A-z]+)"+ 3 * (r"\s+("+RE_FLOAT+")") + "\s*$",
                      name="gradient",
                      repeats=True,
                      startReAction=extract_force
                      )

        return SM(r"\s*ATOM\s+CARTESIAN\s+GRADIENTS\s*$",
                  name="gradients header",
                  subMatchers=[
                      gradient
                  ]
                  )



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
import nomadcore.elements as elements
from turbomoleparser.TurbomoleCommon import RE_FLOAT

logger = logging.getLogger("nomad.turbomoleParser")


class Atom(object):

    def __init__(self, x, y, z, elem, charge):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        self.elem = elem.capitalize()
        self.charge = float(charge)
        self.label = self.elem


class EmbeddingParser(object):

    def __init__(self, context, key="embedding"):
        context[key] = self
        self.__context = context
        self.__backend = None

    def purge_data(self):
        pass

    def set_backend(self, backend):
        self.__backend = backend

    # match builders

    def build_embedding_matcher(self):

        def store_max_multipole(backend, groups):
            index = self.__context.index_system()
            backend.addValue("x_turbomole_pceem_max_multipole", int(groups[0]), index)

        def store_multipole_precision(backend, groups):
            index = self.__context.index_system()
            backend.addValue("x_turbomole_pceem_multipole_precision", float(groups[0]), index)

        def store_cell_separation(backend, groups):
            index = self.__context.index_system()
            backend.addValue("x_turbomole_pceem_min_separation_cells", float(groups[0]), index)

        max_multipole = SM(r"\s*Maximum multipole moment used\s*:\s*([0-9]+)\s*$",
                           name="max multipole",
                           startReAction=store_max_multipole
                           )
        multipole_precision = SM(r"\s*Multipole precision parameter\s*:\s*("+RE_FLOAT+")\s*$",
                                 name="multipole precision",
                                 startReAction=store_multipole_precision
                                 )
        cell_separation = SM(r"\s*Minimum separation between cells\s*:\s*("+RE_FLOAT+")\s*$",
                             name="cell seperation",
                             startReAction=store_cell_separation
                             )
        header = SM(r"\s*\+-+\s*Parameters\s*-*\+\s*$",
                    name="PCEEM parameters",
                    subMatchers=[
                        max_multipole,
                        multipole_precision,
                        cell_separation
                    ]
                    )

        return SM(r"\s*\|\s*EMBEDDING IN PERIODIC POINT CHARGES\s*\|\s*$",
                  name="embedding (PEECM)",
                  subMatchers=[
                      SM(r"\s*\|\s*M. Sierka and A. Burow\s*\|\s*$", name="credits"),
                      header,
                      self.__build_pc_cell_matcher(),
                      self.__build_removed_pc_cluster_matcher(),
                      self.__build_shifted_qm_cluster_matcher()
                  ]
                  )

    def __write_embedding_atom_data(self, gIndex, atoms, linkd_description, add_charges):
        references = {"section_system": self.__context.index_system()}
        index = self.__backend.openSection("section_system_to_system_refs")
        self.__backend.addValue("system_to_system_kind", linkd_description)
        self.__backend.addValue("system_to_system_ref", gIndex)
        self.__backend.setSectionInfo("section_system_to_system_refs", index, references)
        self.__backend.closeSection("section_system_to_system_refs", index)

        pos = np.ndarray(shape=(len(atoms), 3), dtype=float)
        labels = list()
        self.__backend.addValue("number_of_atoms", len(atoms))
        atom_numbers = np.ndarray(shape=(len(atoms),), dtype=float)
        charges = np.ndarray(shape=(len(atoms),), dtype=float)
        for i, atom in enumerate(atoms):
            pos[i, 0:3] = (atom.x, atom.y, atom.z)
            labels.append(atom.elem)
            atom_numbers[i] = elements.get_atom_number(atom.elem)
            charges[i] = atom.charge
        self.__backend.addArrayValues("atom_positions", pos, unit="angstrom")
        self.__backend.addArrayValues("atom_labels", np.asarray(labels, dtype=str))
        self.__backend.addArrayValues("atom_atom_number", atom_numbers)
        if add_charges:
            self.__backend.addArrayValues("x_turbomole_pceem_charges", charges)

    def __build_pc_cell_matcher(self):
        lattice_vectors = list()
        embedding_atoms = list()

        def add_lattice_vector(backend, groups):
            lattice_vectors.append(tuple(float(x) for x in groups[0:3]))

        def add_point_charge(backend, groups):
            new_atom = Atom(x=groups[1], y=groups[2], z=groups[3], elem=groups[0], charge=groups[4])
            embedding_atoms.append(new_atom)

        def finalize(backend, gIndex, section):
            if len(lattice_vectors) != 3:
                logger.error("didn't get expected 3 lattice vectors for PCEEM embedding cell:"
                             + str(len(lattice_vectors)))
            else:
                lattice = np.ndarray(shape=(3, 3))
                for i, vector in enumerate(lattice_vectors[0:3]):
                    lattice[i, :] = vector[:]
                backend.addArrayValues("lattice_vectors", lattice, gIndex, unit="bohr")
            self.__write_embedding_atom_data(gIndex, embedding_atoms,
                                             "periodic point charges for embedding", True)

        lattice_vector = SM(r"\s*("+RE_FLOAT+")"+ 2 * ("\s+("+RE_FLOAT+")")+"\s*$",
                            name="PCEEM lattice",
                            repeats=True,
                            startReAction=add_lattice_vector,
                            required=True
                            )
        point_charge_in_unit_cell = SM(r"\s*([A-z]+)"+4*("\s+("+RE_FLOAT+")")+"\s*$",
                                       name="point charge",
                                       repeats=True,
                                       startReAction=add_point_charge,
                                       required=True)
        point_charge_cell = SM(r"\s*Redefined unit cell content \(au\):\s*$",
                               name="header",
                               subMatchers=[
                                   SM(r"\s*Label\s+Cartesian\s+Coordinates\s+Charge\s*$",
                                      name="header"),
                                   point_charge_in_unit_cell
                               ]
                               )
        return SM(r"\s*Cell vectors \(au\):\s*$",
                  name="PCEEM lattice",
                  sections=["section_system"],
                  required=True,
                  subMatchers=[
                      lattice_vector,
                      point_charge_cell
                  ],
                  onClose={"section_system": finalize}
                  )

    def __build_removed_pc_cluster_matcher(self):
        embedding_atoms = list()

        def add_removed_point_charge(backend, groups):
            new_atom = Atom(x=groups[1], y=groups[2], z=groups[3], elem=groups[0], charge=0)
            embedding_atoms.append(new_atom)

        def finalize(backend, gIndex, section):
            self.__write_embedding_atom_data(gIndex, embedding_atoms,
                                             "removed point charge cluster", False)

        removed_point_charge = SM(r"\s*([A-z]+)"+6*("\s+("+RE_FLOAT+")")+"\s*$",
                                  name="removed point charge",
                                  repeats=True,
                                  startReAction=add_removed_point_charge,
                                  required=True)
        return SM(r"\s*PC cluster transformed to the center of cell 0 \(au\):\s*$",
                  name="header",
                  sections=["section_system"],
                  subMatchers=[
                      SM(r"\s*Label\s+Cartesian\s+Coordinates\s+Cell\s+Indices\s*$", name="header"),
                      removed_point_charge
                  ],
                  onClose={"section_system": finalize}
                  )

    def __build_shifted_qm_cluster_matcher(self):
        embedding_atoms = list()

        def add_shifted_qm_atom(backend, groups):
            new_atom = Atom(x=groups[1], y=groups[2], z=groups[3], elem=groups[0], charge=0)
            embedding_atoms.append(new_atom)

        def finalize(backend, gIndex, section):
            self.__write_embedding_atom_data(gIndex, embedding_atoms,
                                             "shifted embedded QM cluster", False)

        shifted_qm_atom = SM(r"\s*([A-z]+)"+3*("\s+("+RE_FLOAT+")")+"\s*$",
                             name="shifted QM atom",
                             repeats=True,
                             startReAction=add_shifted_qm_atom,
                             required=True)
        return SM(r"\s*QM cluster transformed to the center of cell 0 \(au\):\s*$",
                  name="header",
                  sections=["section_system"],
                  subMatchers=[
                      SM(r"\s*Atom\s+Cartesian\s+Coordinates\s*$", name="header"),
                      shifted_qm_atom
                  ],
                  onClose={"section_system": finalize}
                  )



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nomad.datamodel.encyclopedia import (
    Calculation,
    Properties,
    Energies,
)
from nomad.metainfo import Section
from nomad.normalizing.encyclopedia.context import Context


class PropertiesNormalizer():
    """A base class that is used for processing calculated quantities that
    should be extracted to Encyclopedia.
    """
    def __init__(self, entry_archive, logger):
        self.entry_archive = entry_archive
        self.logger = logger

    def electronic_band_structure(self, properties: Properties, calc_type: str, material_type: str, context: Context, sec_system: Section) -> None:
        """Tries to resolve a reference to a representative electronic band
        structure. Will loop through the available band in reversed order until
        a valid band is found.
        """
        try:
            representative_scc = context.representative_scc
            bands = representative_scc.section_k_band
            if bands is None:
                return

            representative_band = None
            for band in reversed(bands):
                kind = band.band_structure_kind
                if kind != "vibrational":
                    valid = True
                    for segment in band.section_k_band_segment:
                        energies = segment.band_energies
                        k_points = segment.band_k_points
                        labels = segment.band_segm_labels
                        if energies is None or k_points is None or labels is None:
                            valid = False
                    if valid:
                        representative_band = band
                        break

        except Exception:
            return
        if representative_band is not None:
            properties.electronic_band_structure = representative_band.m_path()

            # Add band gap information to metadata if present. The channel with
            # smallest band gap index is chosen as a representative one.
            band_gaps = properties.electronic_band_structure.section_band_gap
            if band_gaps is not None and len(band_gaps) > 0:
                min_gap_index = 0
                min_gap = float("Inf")
                for i, gap in enumerate(band_gaps):
                    value = gap.value
                    if value < min_gap:
                        min_gap_index = i
                        min_gap = value
                representative_gap = band_gaps[min_gap_index]
                bg_value = representative_gap.value
                if bg_value is not None:
                    properties.band_gap = representative_gap.value
                    properties.band_gap_direct = representative_gap.type == "direct"

    def electronic_dos(self, properties: Properties, context: Context) -> None:
        """Tries to resolve a reference to a representative electonic density
        of states. Will not store any data if it cannot be resolved
        unambiguously.
        """
        try:
            representative_scc = context.representative_scc
            doses = representative_scc.section_dos
            if doses is None:
                return

            representative_dos = None
            for dos in reversed(doses):
                kind = dos.dos_kind
                energies = dos.dos_energies
                values = dos.dos_values
                if kind != "vibrational" and energies is not None and values is not None:
                    representative_dos = dos
                    break

        except Exception:
            return
        if representative_dos is not None:
            properties.electronic_dos = representative_dos.m_path()

    def elastic_constants_matrix(self) -> None:
        pass

    def elastic_deformation_energies(self) -> None:
        pass

    def elastic_fitting_parameters(self) -> None:
        pass

    def elastic_moduli(self) -> None:
        pass

    def elastic_properties(self) -> None:
        pass

    def fermi_surface(self) -> None:
        pass

    def has_fermi_surface(self) -> None:
        pass

    def thermodynamical_properties(self, properties: Properties) -> None:
        """Tries to resolve a reference to a representative set of
        thermodynamical properties. Will not store any data if it cannot be
        resolved unambiguously.
        """
        try:
            resolved_section = None
            frame_sequences = self.entry_archive.section_run[0].section_frame_sequence
            for frame_sequence in reversed(frame_sequences):
                thermodynamical_props = frame_sequence.section_thermodynamical_properties
                for thermodynamical_prop in thermodynamical_props:
                    if resolved_section is None:
                        resolved_section = thermodynamical_prop
                    else:
                        self.logger("Could not unambiguously select data to display for specific heat.")
                        return
        except Exception:
            return
        if resolved_section is not None:
            properties.thermodynamical_properties = resolved_section.m_path()

    def phonon_band_structure(self, properties: Properties, context: Context) -> None:
        """Tries to resolve a reference to a representative phonon band
        structure. Will not store any data if it cannot be resolved
        unambiguously.
        """
        try:
            representative_scc = context.representative_scc
            bands = representative_scc.section_k_band
            if bands is None:
                return

            representative_phonon_band = None
            for band in reversed(bands):
                kind = band.band_structure_kind
                if kind == "vibrational":
                    valid = True
                    for segment in band.section_k_band_segment:
                        energies = segment.band_energies
                        k_points = segment.band_k_points
                        labels = segment.band_segm_labels
                        if energies is None or k_points is None or labels is None or "?" in labels:
                            valid = False
                    if valid:
                        representative_phonon_band = band
                        break

        except Exception:
            return
        if representative_phonon_band is not None:
            properties.phonon_band_structure = representative_phonon_band.m_path()

    def phonon_dos(self, properties: Properties, context: Context) -> None:
        """Tries to resolve a reference to a representative phonon density of
        states. Will not store any data if it cannot be resolved unambiguously.
        """
        try:
            representative_scc = context.representative_scc
            doses = representative_scc.section_dos
            if doses is None:
                return

            representative_phonon_dos = None
            for dos in reversed(doses):
                kind = dos.dos_kind
                energies = dos.dos_energies
                values = dos.dos_values
                if kind == "vibrational" and energies is not None and values is not None:
                    if representative_phonon_dos is None:
                        representative_phonon_dos = dos
                    else:
                        self.logger("Could not unambiguously select data to display for phonon density of states.")
                        return

        except Exception:
            return
        if representative_phonon_dos is not None:
            properties.phonon_dos = representative_phonon_dos.m_path()

    def energies(self, properties: Properties, n_atoms: int, representative_scc: Section) -> None:
        if representative_scc is not None:
            energies = Energies()
            energy_found = False
            for energy_name in ["energy_total", "energy_total_T0", "energy_free"]:
                energy_value = getattr(representative_scc, energy_name)
                if energy_value is not None:
                    energy_found = True

                    # The energies are normalized to be per atom
                    setattr(energies, energy_name, energy_value.magnitude / n_atoms)
            if energy_found:
                properties.m_add_sub_section(Properties.energies, energies)

    def normalize(self, context: Context) -> None:
        # There needs to be a valid SCC in order to extract any properties
        representative_scc = context.representative_scc
        if representative_scc is None:
            return

        # Fetch resources
        sec_enc = self.entry_archive.section_metadata.encyclopedia
        properties = sec_enc.properties
        calc_type = context.calc_type
        material_type = context.material_type
        sec_system = context.representative_system
        n_atoms = len(sec_system.atom_labels)

        # Save metainfo
        self.electronic_band_structure(properties, calc_type, material_type, context, sec_system)
        self.electronic_dos(properties, context)
        self.energies(properties, n_atoms, representative_scc)

        # Phonon calculations have a specific set of properties to extract
        if context.calc_type == Calculation.calculation_type.type.phonon_calculation:
            self.thermodynamical_properties(properties)
            self.phonon_band_structure(properties, context)
            self.phonon_dos(properties, context)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import AbsVariable
import logging

logger = logging.getLogger(__name__)


class Variable(AbsVariable):
    """Variables: This command assigns one or more strings to a variable name for evaluation later in the input script or during a simulation.
    - name (string) can only contain alphanumeric characters and underscores
    - value (“string”) can be simple text, it can contain other variables, or it can be a formula.

    equal-style variable VS internal-style variable

    The variables can NOT be re-defined in an input script. There are two exceptions to this rule:
      - First, variables of style string, getenv, internal, equal, vector, atom, and python ARE redefined each time the command is encountered.
      - Second, as described below, if a variable is iterated on to the end of its list of strings via the next command, it is removed from the list of active variables, and is thus available to be re-defined in a subsequent variable command. The delete style does the same thing.

    - the next command can be used to increment which string is assigned to the variable
    - equal: store a formula which when evaluated produces a single numeric value
    - vector: store a formula which produces a vector of such values
    - atom: store a formula which when evaluated produces one numeric value per atom
    - atomfile: can be used anywhere in an input script that atom-style variables are used; they get their per-atom values from a file rather than from a formula
    - python: can be hooked to Python functions using code you provide, so that the variable gets its value from the evaluation of the Python code
    - internal: are used by a few commands which set their value directly
    """
    names = {}

    def __init__(self, *args, **kwargs):
        pass

    def parse(self, name, style, *args):
        pass

    def delete(self, name):
        """delete = no args
        :param name: name of variable to define
        :return:
        """
        if name in self.names:
            del self.names[name]


    def index(self, name, args):
        """index
        :param name: name of variable to define
        :param args: one or more strings
        :return:
        """
        self.names[name] = args

    def loop(self, name, *args):
        """ loop args = N
              N = integer size of loop, loop from 1 to N inclusive
            loop args = N pad
              N = integer size of loop, loop from 1 to N inclusive
              pad = all values will be same length, e.g. 001, 002, ..., 100
            loop args = N1 N2
              N1,N2 = loop from N1 to N2 inclusive
            loop args = N1 N2 pad
              N1,N2 = loop from N1 to N2 inclusive
              pad = all values will be same length, e.g. 050, 051, ..., 100
        :param name: name of variable to define
        :param N: integer size of loop, loop from 1 to N inclusive
        """

        if len(args) == 1:
            self.names[name] = [x for x in range(1, N+1)]
        elif len(args) == 2:
            pass
        elif len(args) == 3:
            pass
        else:
            pass


    def world(self, *args, **kwargs):
        logger.warn("Variable style of {} is not supported yet.".format("word"))
        pass

    def universe(self, *args, **kwargs):
        pass

    def uloop(self, *args, **kwargs):
        pass

    def string(self, *args, **kwargs):
        pass

    def format(self, *args, **kwargs):
        pass

    def getenv(self, *args, **kwargs):
        pass

    def file(self, *args, **kwargs):
        pass

    def atomfile(self, *args, **kwargs):
        pass

    def python(self, *args, **kwargs):
        pass

    def equal(self, *args, **kwargs):
        pass

    def vector(self, *args, **kwargs):
        pass

    def atom(self, *args, **kwargs):
        pass





if __name__ == '__main__':
    pass
    v = Variable()

    str="a b c"
    v.index('a', str.split())

    v.delete('a')

    v.world('a')

    # v.loop('b', 3)


    logger.warn("Not supported yet.")
    print(v.names)



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the DSCF module from TurboMole"""

import logging
import re
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")


class DSCFparser(object):

    def __init__(self, context, key="dscf"):
        context[key] = self
        self.__context = context
        self.__backend = None

    def purge_data(self):
        pass

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{5,}[^+ ]+",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(name="Credits",
                    startReStr=r"\s*d s c f - program\s*$",
                    coverageIgnore=True,
                    subMatchers=[references],
                    endReStr=r"\s*\+-+\+"
                    )

        sub_matchers = [
            header,
            self.__context["method"].build_uhf_matcher(),
            self.__context["geo"].build_qm_geometry_matcher(),
            self.__context["geo"].build_orbital_basis_matcher(),
            self.__context["orbitals"].build_ir_rep_matcher(),
            self.__context["method"].build_dft_functional_matcher(),
            self.__context["embedding"].build_embedding_matcher(),
            self.__context["method"].build_dftd3_vdw_matcher(),
            self.__build_scf_cycle_matcher(),
            self.__context["method"].build_total_energy_matcher(),
            self.__context["orbitals"].build_eigenstate_matcher(),
            Common.build_profiling_matcher(r"\s*dscf profiling\s*$"),
        ]

        return self.__context.build_module_matcher("dscf", sub_matchers, "DSCF",
                                                   self.__context["method"].add_default_functional)

    def __build_scf_cycle_matcher(self):

        class PreviousCycle(object):
            energy = None

        def compute_energy_difference(backend, groups):
            backend.addRealValue("energy_total_scf_iteration", float(groups[0]), unit="hartree")
            backend.addRealValue("x_turbomole_energy_1electron_scf_iteration",
                                 float(groups[1]), unit="hartree")
            backend.addRealValue("x_turbomole_energy_2electron_scf_iteration",
                                 float(groups[2]), unit="hartree")
            # TODO: clarify the meaning of the "NORM[dD(SAO)]" and "TOL" columns in the output
            if PreviousCycle.energy:
                backend.addRealValue("energy_change_scf_iteration",
                                 float(groups[0]) - PreviousCycle.energy, unit="hartree")
            PreviousCycle.energy = float(groups[0])

        def store_damping(backend, groups):
            backend.addRealValue("x_turbomole_damping_scf_iteration",
                                 1.0 / (1.0 + float(groups[0])))

        total_energy_matcher = SM(r"\s*[0-9]+\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")"
                                  "\s+("+RE_FLOAT+")\s+("+RE_FLOAT+")",
                                  startReAction=compute_energy_difference,
                                  name="SCF E total",
                                  required=True
                                  )
        xc_energy_matcher = SM(r"\s*Exc =\s*(?P<energy_XC_scf_iteration__hartree>"+RE_FLOAT+")"
                               r"\s+N =\s*("+RE_FLOAT+")",
                               name="SCF E xc"
                               )

        # TODO: identify the units of these norms
        norm_diis = SM(r"\s*Norm of current diis error:\s*(?P<x_turbomole_norm_diis_scf_iteration>"
                       + RE_FLOAT+r")\s*$",
                       name="SCF norm DIIS"
                       )
        norm_fia_block = SM(r"\s*max. resid. norm for Fia\-block\s*=\s*(?P"
                            r"<x_turbomole_norm_fia_scf_iteration>"+RE_FLOAT+")\s*for orbital\s+"
                            r"(?P<x_turbomole_norm_fia_orbital_scf_iteration>[0-9]+[a-z][0-9'\"]?"
                            r"\s*(?:alpha|beta)?)\s*$",
                            name="SCF norm Fia"
                            )
        norm_fock = SM(r"\s*max. resid. fock norm\s*=\s*(?P"
                       r"<x_turbomole_norm_fock_scf_iteration>"+RE_FLOAT+")\s*for orbital\s+"
                       r"(?P<x_turbomole_norm_fock_orbital_scf_iteration>[0-9]+[a-z][0-9'\"]?"
                       r"\s*(?:alpha|beta)?)\s*$",
                       name="SCF norm Fock"
                       )
        eigenval_change = SM(r"\s*Delta Eig\.\s*=\s*(?P<x_turbomole_delta_eigenvalues__eV>"
                             +RE_FLOAT+r")\s+eV\s*$",
                             name="SCF D eigenvals")

        scf_iteration = SM(r"\s*current damping\s*:\s*("+RE_FLOAT+")\s*$",
                           name="SCF damping",
                           repeats=True,
                           sections=["section_scf_iteration"],
                           startReAction=store_damping,
                           subMatchers=[
                               SM("\s*ITERATION\s+ENERGY\s+1e-ENERGY\s+2e-ENERGY\s+"
                                  "NORM\[dD\(SAO\)\]\s+TOL",
                                  name="SCF iteration",
                                  required=True
                                  ),
                               total_energy_matcher,
                               xc_energy_matcher,
                               norm_diis,
                               norm_fia_block,
                               norm_fock,
                               eigenval_change
                           ]
                           )

        return SM(r"\s*scf convergence criterion : increment of total energy <\s*"+RE_FLOAT+"\s*$",
                  name="HF/DFT SCF",
                  required=True,
                  subMatchers=[
                      SM("\s*time elapsed for pre-SCF steps : cpu\s+([0-9]+\.[0-9]+)\s+sec",
                         name="SCF preparation"),
                      SM("\s*wall\s+([0-9]+\.[0-9]+)\s+sec",
                         name="SCF preparation"),
                      scf_iteration
                  ]
                  )



import lark
from lark import v_args
from elasticsearch_dsl import Q, Text, Keyword, Integer, Field
from optimade.models import CHEMICAL_SYMBOLS, ATOMIC_NUMBERS


_cmp_operators = {">": "gt", ">=": "gte", "<": "lt", "<=": "lte"}
_rev_cmp_operators = {">": "<", ">=": "<=", "<": ">", "<=": "=>"}
_has_operators = {"ALL": "must", "ANY": "should"}
_length_quantities = {
    "elements": "nelements",
    "elements_rations": "nelements",
    "dimension_types": "dimension_types",
}


class Quantity:
    """ Class to provide information about available quantities to the transformer.

    The elasticsearch transformer will :class:`Quantity`s to (a) do some semantic checks,
    (b) map quantities to the underlying elastic index.

    Attributes:
        name: The name of the quantity as used in the filter expressions.
        es_field: The name of the field for this quanity in elastic search, will be
            ``name`` by default.
        elastic_mapping_type: A decendent of an elasticsearch_dsl Field that denotes which
            mapping was used in the elastic search index.
        length_quantity: Elasticsearch does not support length of arrays, but we can
            map fields with array to other fields with ints about the array length. The
            LENGTH operator will only be supported for quantities with this attribute.
        has_only_quantity: Elasticsearch does not support exclusive search on arrays, like
            a list of chemical elements. But, we can order all elements by atomic number
            and use a keyword field with all elements to perform this search. This only
            works for elements (i.e. labels in ``CHEMICAL_SYMBOLS``) and quantities
            with this attribute.
        nested_quantity: To support optimade's 'zipped tuple' feature (e.g.
            'elements:elements_ratios HAS "H":>0.33), we use elasticsearch nested objects
            and nested queries. This quantity will provide the field for the nested
            object that contains the quantity (and others). The zipped tuples will only
            work for quantities that share the same nested object quantity.
    """

    def __init__(
        self,
        name,
        es_field: str = None,
        elastic_mapping_type: Field = None,
        length_quantity: "Quantity" = None,
        has_only_quantity: "Quantity" = None,
        nested_quantity: "Quantity" = None,
    ):

        self.name = name
        self.es_field = es_field if es_field is not None else name
        self.elastic_mapping_type = (
            Keyword if elastic_mapping_type is None else elastic_mapping_type
        )
        self.length_quantity = length_quantity
        self.has_only_quantity = has_only_quantity
        self.nested_quantity = nested_quantity

    def __repr__(self):
        return self.name


class ElasticTransformer(lark.Transformer):
    """ Transformer that transforms ``v0.10.1`` grammar parse trees into queries.

    Uses elasticsearch_dsl and will produce a :class:`Q` instance.

    Arguments:
        quantities: A list of :class:`Quantity`s that describe how optimade (and other)
            quantities are mapped to the elasticsearch index.
    """

    def __init__(self, quantities):
        self.index_mapping = {quantity.name: quantity for quantity in quantities}

    def _field(self, quantity, nested=None):
        if nested is not None:
            return "%s.%s" % (nested.es_field, quantity.name)

        return quantity.es_field

    def _query_op(self, quantity, op, value, nested=None):
        """
        Return a range, match, or term query for the given quantity, comparison
        operator, and value
        """
        field = self._field(quantity, nested=nested)
        if op in _cmp_operators:
            return Q("range", **{field: {_cmp_operators[op]: value}})

        if quantity.elastic_mapping_type == Text:
            query_type = "match"
        elif quantity.elastic_mapping_type in [Keyword, Integer]:
            query_type = "term"
        else:
            raise NotImplementedError("Quantity has unsupported ES field type")

        if op in ["=", ""]:
            return Q(query_type, **{field: value})

        if op == "!=":
            return ~Q(  # pylint: disable=invalid-unary-operand-type
                query_type, **{field: value}
            )

    def _has_query_op(self, quantities, op, predicate_zip_list):
        """
        Returns a bool query that combines the operator queries ():func:`_query_op`)
        for each predicate and zipped quantity pericates combinations.
        """
        if op == "HAS":
            kind = "must"  # in case of HAS we do a must over the "list" of the one given element
        elif op == "HAS ALL":
            kind = "must"
        elif op == "HAS ANY":
            kind = "should"
        elif op == "HAS ONLY":
            # HAS ONLY comes with heavy limitations, because there is no such thing
            # in elastic search. Only supported for elements, where we can construct
            # an anonymous "formula" based on elements sorted by order number and
            # can do a = comparision to check if all elements are contained
            if len(quantities) > 1:
                raise Exception("HAS ONLY is not supported with zip")
            quantity = quantities[0]

            if quantity.has_only_quantity is None:
                raise Exception("HAS ONLY is not supported by %s" % quantity.name)

            def values():
                for predicates in predicate_zip_list:
                    if len(predicates) != 1:
                        raise Exception("Tuples not supported in HAS ONLY")
                    op, value = predicates[0]
                    if op != "=":
                        raise Exception("Predicated not supported in HAS ONLY")
                    if not isinstance(value, str):
                        raise Exception("Only strings supported in HAS ONLY")
                    yield value

            try:
                order_numbers = list([ATOMIC_NUMBERS[element] for element in values()])
                order_numbers.sort()
                value = "".join(
                    [CHEMICAL_SYMBOLS[number - 1] for number in order_numbers]
                )
            except KeyError:
                raise Exception("HAS ONLY is only supported for chemical symbols")

            return Q("term", **{quantity.has_only_quantity.name: value})
        else:
            raise NotImplementedError

        queries = [
            self._has_query(quantities, predicates) for predicates in predicate_zip_list
        ]
        return Q("bool", **{kind: queries})

    def _has_query(self, quantities, predicates):
        """
        Returns a bool query that combines the operator queries ():func:`_query_op`)
        for quantity pericate combination.
        """
        if len(quantities) != len(predicates):
            raise Exception(
                "Tuple length does not match: %s <o> %s "
                % (":".join(quantities), ":".join(predicates))
            )

        if len(quantities) == 1:
            o, value = predicates[0]
            return self._query_op(quantities[0], o, value)

        nested_quantity = quantities[0].nested_quantity
        same_nested_quantity = any(
            q.nested_quantity != nested_quantity for q in quantities
        )
        if nested_quantity is None or same_nested_quantity:
            raise Exception(
                "Expression with tuples are only supported for %s"
                % ", ".join(quantities)
            )

        queries = [
            self._query_op(quantity, o, value, nested=nested_quantity)
            for quantity, (o, value) in zip(quantities, predicates)
        ]

        return Q(
            "nested",
            path=self._field(nested_quantity),
            query=dict(bool=dict(must=queries)),
        )

    def __default__(self, tree, children, *args, **kwargs):
        """ Default behavior for rules that only replace one symbol with another """
        return children[0]

    def filter(self, args):
        # filter: expression*
        if len(args) == 1:
            return args[0]
        return Q("bool", **{"must": args})

    def expression_clause(self, args):
        # expression_clause: expression_phrase ( _AND expression_phrase )*
        result = args[0]
        for arg in args[1:]:
            result &= arg
        return result

    def expression(self, args):
        # expression: expression_clause ( _OR expression_clause )*
        result = args[0]
        for arg in args[1:]:
            result |= arg
        return result

    def expression_phrase(self, args):
        # expression_phrase: [ NOT ] ( operator | "(" expression ")" )
        if args[0] == "NOT":
            return ~args[1]
        return args[0]

    @v_args(inline=True)
    def property_first_comparison(self, quantity, query):
        # property_first_comparison: property *_rhs
        return query(quantity)

    @v_args(inline=True)
    def constant_first_comparison(self, value, op, quantity):
        # constant_first_comparison: constant OPERATOR ( non_string_value | ...not_implemented_string )
        if not isinstance(quantity, Quantity):
            raise Exception("Only quantities can be compared to constant values.")

        return self._query_op(quantity, _rev_cmp_operators[op], value)

    @v_args(inline=True)
    def value_op_rhs(self, op, value):
        # value_op_rhs: OPERATOR value
        return lambda quantity: self._query_op(quantity, op, value)

    def length_op_rhs(self, args):
        # length_op_rhs: LENGTH [ OPERATOR ] signed_int
        value = args[-1]
        if len(args) == 3:
            op = args[1]
        else:
            op = "="

        def query(quantity):
            if quantity.length_quantity is None:
                raise Exception("LENGTH is not supported for %s" % quantity.name)
            quantity = quantity.length_quantity
            return self._query_op(quantity, op, value)

        return query

    @v_args(inline=True)
    def known_op_rhs(self, _, value):
        # known_op_rhs: IS ( KNOWN | UNKNOWN )

        def query(quantity):
            query = Q("exists", field=self._field(quantity))
            if value == "KNOWN":
                return query
            elif value == "UNKNOWN":
                return ~query  # pylint: disable=invalid-unary-operand-type
            raise NotImplementedError

        return query

    def set_op_rhs(self, args):
        # set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ... )
        values = args[-1]
        if not isinstance(values, list):
            if len(args) == 3:
                op = args[1]
            else:
                op = "="
            values = [(op, values)]

        if len(args) == 3:
            op = "HAS " + args[1]
        else:
            op = "HAS"

        def query(quantity):
            return self._has_query_op([quantity], op, [[value] for value in values])

        return query

    def set_zip_op_rhs(self, args):
        # set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list | ANY value_zip_list )
        add_on = args[0]
        values = args[-1]
        if len(args) == 4:
            op = "HAS " + args[2]
        else:
            op = "HAS"
            values = [values]

        return lambda quantity: self._has_query_op([quantity] + add_on, op, values)

    def property_zip_addon(self, args):
        return args

    def value_zip(self, args):
        return self.value_list(args)

    def value_zip_list(self, args):
        return args

    def value_list(self, args):
        result = []
        op = "="
        for arg in args:
            if arg in ["<", "<=", ">", ">=", "!=", "="]:
                op = arg
            else:
                result.append((op, arg,))
                op = "="
        return result

    def fuzzy_string_op_rhs(self, args):
        op = args[0]
        value = args[-1]
        if op == "CONTAINS":
            wildcard = "*%s*" % value
        if op == "STARTS":
            wildcard = "%s*" % value
        if op == "ENDS":
            wildcard = "*%s" % value

        return lambda quantity: Q("wildcard", **{self._field(quantity): wildcard})

    def property(self, args):
        # property: IDENTIFIER ( "." IDENTIFIER )*
        quantity_name = args[0]

        if quantity_name not in self.index_mapping:
            raise Exception("%s is not a searchable quantity" % quantity_name)

        quantity = self.index_mapping.get(quantity_name, None)
        if quantity is None:
            quantity = Quantity(name=quantity_name)

        return quantity

    @v_args(inline=True)
    def string(self, string):
        # string: ESCAPED_STRING
        return string.strip('"')

    @v_args(inline=True)
    def signed_int(self, number):
        # signed_int : SIGNED_INT
        return int(number)

    @v_args(inline=True)
    def number(self, number):
        # number: SIGNED_INT | SIGNED_FLOAT
        if number.type == "SIGNED_INT":
            type_ = int
        elif number.type == "SIGNED_FLOAT":
            type_ = float
        return type_(number)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from LAMMPS.BaseClasses import Units

logger = logging.getLogger(__name__)


# """
# The units command also sets the timestep size and neighbor skin distance to default values for each style:
# For style lj these are dt = 0.005 tau and skin = 0.3 sigma.
# For style real these are dt = 1.0 fmsec and skin = 2.0 Angstroms.
# For style metal these are dt = 0.001 psec and skin = 2.0 Angstroms.
# For style si these are dt = 1.0e-8 sec and skin = 0.001 meters.
# For style cgs these are dt = 1.0e-8 sec and skin = 0.1 cm.
# For style electron these are dt = 0.001 fmsec and skin = 2.0 Bohr.
# For style micro these are dt = 2.0 microsec and skin = 0.1 micrometers.
# For style nano these are dt = 0.00045 nanosec and skin = 0.1 nanometers.
# """

def units(style, parent=None):
    """Wrapper for unit style command"""
    if not parent:
        print("hehe")


class _units(object):
    # According Lammps soucre code the physical constants from:
    # http://physics.nist.gov/cuu/Constants/Table/allascii.txt
    # using thermochemical calorie = 4.184 J

    # constants
    mole        = 6.022140857e23    # Avogadro constant [mol^-1]
    Kb          = 1.38064852e-23    # Boltzmann constant [J K^-1]
    perm0       = 8.854187817e-12   # vacuum permittivity / electric constant [F m^-1]

    # mass [kilograms]
    kilogram    = 1.
    gram        = 1e-3
    picogram    = 1e-15
    attogram    = 1e-21
    atomicmass  = 1.660539040e-27   # atomic mass units [kg]

    # distance [meters]
    meter       = 1.
    cm          = 1e-2
    micrometer  = 1e-6
    nanometer   = 1e-9
    Angstrom    = 1.00001495e-10
    Bohr        = 0.52917721067e-10 # Bohr radius [m]

    # time [seconds]
    second      = 1.
    microsecond = 1e-6
    nanosecond  = 1e-9
    picosecond  = 1e-12
    femtosecond = 1e-15
    atomictime  = 2.418884326509e-17  # 1.03275e-15       # atomic time units

    # energy [Joules]
    Joule       = 1.
    Kcal        = 4.184             # Thermochemical calorie
    ergs        = 1e-7
    eV          = 1.6021766208e-19  # electron volt [J]
    Hartree     = 4.359744650e-18   # Hartree energy [J]


    # velocity [meters/second]


    # force [Newtons]
    Newton      = 1.
    dyne        = 1e-5

    # torque [Newton-meters]

    # temperature [Kelvin]
    Kelvin      = 1.

    # pressure [Pascals]
    Pascal      = 1.
    atmosphere  = 101325.
    bar         = 1e5

    # dynamic_viscosity [Pascal*second]
    Poise       = 0.1

    # charge [Coulombs (1.6021765e-19 is a proton)]
    Coulomb     = 1.
    picocoulomb = 1e-6
    charge_e    = 1.6021766208e-19
    statcoulomb = 1/2997924580      # statcoulomb (statC) or franklin (Fr) or electrostatic unit of charge (esu)

    # dipole [Coulombs*meters]
    # electric_field [volts/meter]
    volt        = 1.
    statvolt    = 299.792458
    Debye       = 1/299792458 * 1e-21

    # density [kilograms/meter^dim]





class lj(Units):
    """For style lj, all quantities are unitless. Without loss of generality, LAMMPS sets the fundamental quantities mass, sigma, epsilon, and the Boltzmann constant = 1. The masses, distances, energies you specify are multiples of these fundamental values. The formulas relating the reduced or unitless quantity (with an asterisk) to the same quantity with units is also given. Thus you can use the mass & sigma & epsilon values for a specific material and convert the results from a unitless LJ simulation into physical quantities.
        - mass = mass or m
        - distance = sigma, where x* = x / sigma
        - time = tau, where t* = t (epsilon / m / sigma^2)^1/2
        - energy = epsilon, where E* = E / epsilon
        - velocity = sigma/tau, where v* = v tau / sigma
        - force = epsilon/sigma, where f* = f sigma / epsilon
        - torque = epsilon, where t* = t / epsilon
        - temperature = reduced LJ temperature, where T* = T Kb / epsilon
        - pressure = reduced LJ pressure, where P* = P sigma^3 / epsilon
        - dynamic viscosity = reduced LJ viscosity, where eta* = eta sigma^3 / epsilon / tau
        - charge = reduced LJ charge, where q* = q / (4 pi perm0 sigma epsilon)^1/2
        - dipole = reduced LJ dipole, moment where *mu = mu / (4 pi perm0 sigma^3 epsilon)^1/2
        - electric field = force/charge, where E* = E (4 pi perm0 sigma epsilon)^1/2 sigma / epsilon
        - density = mass/volume, where rho* = rho sigma^dim"""

    style = "lj"

    def __init__(self, mass=1., sigma=1., epsilon=1.):
        self._mass = mass
        self._sigma = sigma
        self._epsilon = epsilon

    @property
    def _tau(self):
        from math import sqrt
        return self._sigma * sqrt(self.mass/self._epsilon)

    @property
    def mass(self):
        return self._mass

    @property
    def distance(self):
        return self._sigma

    @property
    def time(self):
        return self._tau

    @property
    def energy(self):
        return self._epsilon

    @property
    def velocity(self):
        return self._sigma/self._tau

    @property
    def force(self):
        return self._epsilon/self._sigma

    @property
    def torque(self):
        return self._epsilon

    @property
    def temperature(self):
        return self._epsilon/_units.Kb

    @property
    def pressure(self):
        return self._epsilon/self._sigma**3

    @property
    def dynamic_viscosity(self):
        return self._tau * self._epsilon / self._sigma**3

    @property
    def charge(self):
        from math import pi, sqrt
        return sqrt(4 * pi * _units.perm0 * self._sigma * self._epsilon)

    @property
    def dipole(self):
        from math import pi, sqrt
        return sqrt(4 * pi * _units.perm0 * self._sigma**3 * self._epsilon)

    @property
    def electric_field(self):
        from math import pi, sqrt
        return self._epsilon / self._sigma / sqrt(4 * pi * _units.perm0 * self._sigma * self._epsilon)

    @property
    def density(self, dim=3):
        return self._mass / self._sigma**dim

    def parse(selfs):
        logger.warn("NOT SUPPORTED: For LJ style all quantities are unitless")

class real(Units):
    """For style real, these are the units:
        - mass = grams/mole
        - distance = Angstroms
        - time = femtoseconds
        - energy = Kcal/mole
        - velocity = Angstroms/femtosecond
        - force = Kcal/mole-Angstrom
        - torque = Kcal/mole
        - temperature = Kelvin
        - pressure = atmospheres
        - dynamic viscosity = Poise
        - charge = multiple of electron charge (1.0 is a proton)
        - dipole = charge*Angstroms
        - electric field = volts/Angstrom
        - density = gram/cm^dim"""

    style = "real"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.gram/_units.mole

    @property
    def distance(self):
        return _units.Angstrom

    @property
    def time(self):
        return _units.femtosecond

    @property
    def energy(self):
        return _units.Kcal / _units.mole

    @property
    def velocity(self):
        return _units.Angstrom / _units.femtosecond

    @property
    def force(self):
        return _units.Kcal / _units.mole / _units.Angstrom

    @property
    def torque(self):
        return _units.Kcal / _units.mole

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.atmosphere

    @property
    def dynamic_viscosity(self):
        return _units.Poise

    @property
    def charge(self):
        return _units.charge_e

    @property
    def dipole(self):
        return _units.charge_e * _units.Angstrom

    @property
    def electric_field(self):
        return _units.volt / _units.Angstrom

    @property
    def density(self, dim=3):
        return _units.gram / _units.cm**dim

class metal(Units):
    """For style metal, these are the units:
        - mass = grams/mole
        - distance = Angstroms
        - time = picoseconds
        - energy = eV
        - velocity = Angstroms/picosecond
        - force = eV/Angstrom
        - torque = eV
        - temperature = Kelvin
        - pressure = bars
        - dynamic viscosity = Poise
        - charge = multiple of electron charge (1.0 is a proton)
        - dipole = charge*Angstroms
        - electric field = volts/Angstrom
        - density = gram/cm^dim"""

    style = "metal"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.gram/_units.mole

    @property
    def distance(self):
        return _units.Angstrom

    @property
    def time(self):
        return _units.picosecond

    @property
    def energy(self):
        return _units.eV

    @property
    def velocity(self):
        return _units.Angstrom / _units.picosecond

    @property
    def force(self):
        return _units.eV / _units.Angstrom

    @property
    def torque(self):
        return _units.eV

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.bar

    @property
    def dynamic_viscosity(self):
        return _units.Poise

    @property
    def charge(self):
        return _units.charge_e

    @property
    def dipole(self):
        return _units.charge_e * _units.Angstrom

    @property
    def electric_field(self):
        return _units.volt / _units.Angstrom

    @property
    def density(self, dim=3):
        return _units.gram / _units.cm**dim


class si(Units):
    """For style si, these are the units:
        - mass = kilograms
        - distance = meters
        - time = seconds
        - energy = Joules
        - velocity = meters/second
        - force = Newtons
        - torque = Newton-meters
        - temperature = Kelvin
        - pressure = Pascals
        - dynamic viscosity = Pascal*second
        - charge = Coulombs (1.6021765e-19 is a proton)
        - dipole = Coulombs*meters
        - electric field = volts/meter
        - density = kilograms/meter^dim"""

    style = "si"

    def __init__(self):
        pass


    @property
    def mass(self):
        return _units.kilogram

    @property
    def distance(self):
        return _units.meter

    @property
    def time(self):
        return _units.second

    @property
    def energy(self):
        return _units.Joule

    @property
    def velocity(self):
        return _units.meter / _units.second

    @property
    def force(self):
        return _units.Newton

    @property
    def torque(self):
        return _units.Newton * _units.meter

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.Pascal

    @property
    def dynamic_viscosity(self):
        return _units.Pascal * _units.second

    @property
    def charge(self):
        return _units.Coulomb

    @property
    def dipole(self):
        return _units.Coulomb * _units.meter

    @property
    def electric_field(self):
        return _units.volt / _units.meter

    @property
    def density(self, dim=3):
        return _units.kilogram / _units.meter**dim


class cgs(Units):
    """For style cgs, these are the units:
        - mass = grams
        - distance = centimeters
        - time = seconds
        - energy = ergs
        - velocity = centimeters/second
        - force = dynes
        - torque = dyne-centimeters
        - temperature = Kelvin
        - pressure = dyne/cm^2 or barye = 1.0e-6 bars
        - dynamic viscosity = Poise
        - charge = statcoulombs or esu (4.8032044e-10 is a proton)
        - dipole = statcoul-cm = 10^18 debye
        - electric field = statvolt/cm or dyne/esu
        - density = grams/cm^dim"""

    style = "cgs"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.gram

    @property
    def distance(self):
        return _units.cm

    @property
    def time(self):
        return _units.second

    @property
    def energy(self):
        return _units.ergs

    @property
    def velocity(self):
        return _units.cm/_units.second

    @property
    def force(self):
        return _units.dyne

    @property
    def torque(self):
        return _units.dyne * _units.cm

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.dyne/_units.cm**2

    @property
    def dynamic_viscosity(self):
        return _units.Poise

    @property
    def charge(self):
        return _units.statcoulomb

    @property
    def dipole(self):
        return _units.statcoulomb * _units.cm

    @property
    def electric_field(self):
        return _units.statvolt/_units.cm

    @property
    def density(self, dim=3):
        return _units.gram/_units.cm**dim

class electron(Units):
    """For style electron, these are the units:
        - mass = atomic mass units
        - distance = Bohr
        - time = femtoseconds
        - energy = Hartrees
        - velocity = Bohr/atomic time units [1.03275e-15 seconds]
        - force = Hartrees/Bohr
        - temperature = Kelvin
        - pressure = Pascals
        - charge = multiple of electron charge (1.0 is a proton)
        - dipole moment = Debye
        - electric field = volts/cm"""

    style = "electron"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.atomicmass

    @property
    def distance(self):
        return _units.Bohr

    @property
    def time(self):
        return _units.femtosecond

    @property
    def energy(self):
        return _units.Hartree

    @property
    def velocity(self):
        return _units.Bohr / _units.atomictime

    @property
    def force(self):
        return _units.Hartree / _units.Bohr

    @property
    def torque(self):
        return _units.Hartree

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.Pascal

    @property
    def dynamic_viscosity(self):
        return _units.Poise

    @property
    def charge(self):
        return _units.charge_e

    @property
    def dipole(self):
        return _units.Debye

    @property
    def electric_field(self):
        return _units.volt / _units.cm

    @property
    def density(self, dim=3):
        return _units.gram / _units.cm**dim

class micro(Units):
    """For style micro, these are the units:
        - mass = picograms
        - distance = micrometers
        - time = microseconds
        - energy = picogram-micrometer^2/microsecond^2
        - velocity = micrometers/microsecond
        - force = picogram-micrometer/microsecond^2
        - torque = picogram-micrometer^2/microsecond^2
        - temperature = Kelvin
        - pressure = picogram/(micrometer-microsecond^2)
        - dynamic viscosity = picogram/(micrometer-microsecond)
        - charge = picocoulombs (1.6021765e-7 is a proton)
        - dipole = picocoulomb-micrometer
        - electric field = volt/micrometer
        - density = picograms/micrometer^dim"""

    style = "micro"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.picogram

    @property
    def distance(self):
        return _units.micrometer

    @property
    def time(self):
        return _units.microsecond

    @property
    def energy(self):
        return _units.picogram * _units.micrometer**2 / _units.microsecond**2

    @property
    def velocity(self):
        return _units.micrometer / _units.microsecond

    @property
    def force(self):
        return _units.picogram * _units.micrometer / _units.microsecond**2

    @property
    def torque(self):
        return _units.picogram * _units.micrometer**2 / _units.microsecond**2

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.picogram / (_units.micrometer * _units.microsecond**2)

    @property
    def dynamic_viscosity(self):
        return _units.picogram / (_units.micrometer * _units.microsecond)

    @property
    def charge(self):
        return _units.picocoulomb

    @property
    def dipole(self):
        return _units.picocoulomb * _units.micrometer

    @property
    def electric_field(self):
        return _units.volt / _units.micrometer

    @property
    def density(self, dim=3):
        return _units.picogram / _units.micrometer**dim

class nano(Units):
    """For style nano, these are the units:
        - mass = attograms
        - distance = nanometers
        - time = nanoseconds
        - energy = attogram-nanometer^2/nanosecond^2
        - velocity = nanometers/nanosecond
        - force = attogram-nanometer/nanosecond^2
        - torque = attogram-nanometer^2/nanosecond^2
        - temperature = Kelvin
        - pressure = attogram/(nanometer-nanosecond^2)
        - dynamic viscosity = attogram/(nanometer-nanosecond)
        - charge = multiple of electron charge (1.0 is a proton)
        - dipole = charge-nanometer
        - electric field = volt/nanometer
        - density = attograms/nanometer^dim"""

    style = "nano"

    def __init__(self):
        pass

    @property
    def mass(self):
        return _units.attogram

    @property
    def distance(self):
        return _units.nanometer

    @property
    def time(self):
        return _units.nanometer

    @property
    def energy(self):
        return _units.attogram * _units.nanometer**2 / _units.nanosecond**2

    @property
    def velocity(self):
        return _units.nanometer / _units.nanosecond

    @property
    def force(self):
        return _units.attogram * _units.nanometer / _units.nanosecond**2

    @property
    def torque(self):
        return _units.attogram * _units.nanometer**2 / _units.nanosecond**2

    @property
    def temperature(self):
        return _units.Kelvin

    @property
    def pressure(self):
        return _units.attogram / (_units.nanometer * _units.nanosecond**2)

    @property
    def dynamic_viscosity(self):
        return _units.attogram / (_units.nanometer * _units.nanosecond)

    @property
    def charge(self):
        return _units.charge_e

    @property
    def dipole(self):
        return _units.charge_e * _units.nanometer

    @property
    def electric_field(self):
        return _units.volt / _units.nanometer

    @property
    def density(self, dim=3):
        return _units.attogram / _units.nanometer**dim


if __name__ == '__main__':

    # u = lj()
    # u.parse()
    # print(u)
    # print(u.info())

    styles = ['lj', 'real', 'metal', 'si', 'cgs', 'electron', 'micro', 'nano']

    for style in styles:
        u = eval(style)()
        print(u)
        print(u.info())




# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Blueprint
from flask_restplus import Api
import urllib.parse

from nomad import config

blueprint = Blueprint('optimade', __name__)

base_url = 'https://%s/%s/optimade' % (
    config.services.api_host.strip('/'),
    config.services.api_base_path.strip('/'))


def url(endpoint: str = None, version='v1', prefix=None, **kwargs):
    ''' Returns the full optimade api url (for a given endpoint) including query parameters. '''
    if endpoint is not None:
        url = '/' + endpoint
    else:
        url = ''

    if version is not None:
        url = '/' + version + url

    if prefix is not None:
        url = '/' + prefix + url

    url = base_url + url

    if len(kwargs) > 0:
        return '%s?%s' % (url, urllib.parse.urlencode(kwargs))
    else:
        return url


api = Api(
    blueprint,
    version='1.0', title='NOMAD\'s OPTiMaDe API implementation',
    description='NOMAD\'s OPTiMaDe API implementation, version 1.0.0.',
    validate=True)
''' Provides the flask restplust api instance for the optimade api'''


# For some unknown reason it is necessary for each fr api to have a handler.
# Otherwise the global app error handler won't be called.
@api.errorhandler(Exception)
def errorhandler(error):
    '''When an internal server error is caused by an unexpected exception.'''
    return str(error)



import numpy as np
import logging
import re
logger = logging.getLogger(__name__)


#===============================================================================
def iread(filepath, columns=None, delimiter=r"\s+", comments=r"#", start=None, end=None, n_conf=None):
    """Used to iterate a CSV-like file. If a separator is provided the file
    is iterated one configuration at a time. Only keeps one configuration
    of the file in memory. If no separator is given, the whole file will be
    handled.

    The contents are separated into configurations whenever the separator
    regex is encountered on a line.

    Args:
        filepath: Path to the CSV like file to be processed.
        columns: List of integers indicating the columns of interest in the CSV file.
        start: A regex that is used to indicate the start of a new configuration.
        end: A regex that is used to indicate the end of a configuration.
        comments: A regex that is used identify comments in the file that are ignored.
        n_conf: Number of lines in a configuration. If you want to use multiple
            lines as a single configuration.
    """

    def split_line(line):
        """Chop off comments, strip, and split at delimiter.
        """
        if line.isspace():
            return None
        if comments:
            line = compiled_comments.split(line, maxsplit=1)[0]
        line = line.strip('\r\n ')
        if line:
            return compiled_delimiter.split(line)
        else:
            return None

    def is_end(line):
        """Check if the given line matches the separator pattern.
        Separators are used to split a file into multiple configurations.
        """
        if end:
            return compiled_end.search(line)
        return False

    def is_start(line):
        """Check if the given line matches the separator pattern.
        Separators are used to split a file into multiple configurations.
        """
        if start:
            return compiled_start.search(line)
        return False

    # Precompile the different regexs before looping
    compiled_delimiter = re.compile(delimiter)
    if comments:
        comments = (re.escape(comment) for comment in comments)
        compiled_comments = re.compile('|'.join(comments))
    if end:
        compiled_end = re.compile(end)
    if start:
        compiled_start = re.compile(start)

    # Columns as list
    if columns is not None:
        columns = list(columns)

    # Start iterating
    configuration = []
    started = False

    # If no starting and ending condition are provided, read configuration by line
    if start is None and end is None and n_conf is None:
        with open(filepath, "r") as f:
            for line in f:  # This actually reads line by line and only keeps the current line in memory
                # Ignore comments, separate by delimiter
                vals = split_line(line)
                line_forces = []
                if vals:
                    for column in columns:
                        try:
                            value = vals[column]
                        except IndexError:
                            logger.warning("The given index '{}' could not be found on the line '{}'. The given delimiter or index could be wrong.".format(column, line))
                            return
                        try:
                            value = float(value)
                        except ValueError:
                            logger.warning("Could not cast value '{}' to float. Currently only floating point values are accepted".format(value))
                            return
                        else:
                            line_forces.append(value)
                    yield np.array(line_forces)

    # If starting and ending condition are provided, after starting condition
    # is detected, add the values from lines to a new array that is returned
    # when the end condition is met
    elif start is not None and end is not None:
        with open(filepath, "r") as f:
            for line in f:  # This actually reads line by line and only keeps the current line in memory

                # If a start regex is provided, use it to detect the start of a configuration
                if is_start(line):
                    started = True
                    continue

                # If separator encountered, yield the stored configuration
                if is_end(line):
                    started = False
                    if configuration:
                        yield np.array(configuration)
                        configuration = []

                elif start is not None and started:
                    # Ignore comments, separate by delimiter
                    vals = split_line(line)
                    line_forces = []
                    if vals:
                        for column in columns:
                            try:
                                value = vals[column]
                            except IndexError:
                                logger.warning("The given index '{}' could not be found on the line '{}'. The given delimiter or index could be wrong.".format(column, line))
                                return
                            try:
                                value = float(value)
                            except ValueError:
                                logger.warning("Could not cast value '{}' to float. Currently only floating point values are accepted".format(value))
                                return
                            else:
                                line_forces.append(value)
                        configuration.append(line_forces)

            # The last configuration is yielded even if separator is not present at
            # the end of file or is not given at all
            if configuration:
                yield np.array(configuration)

    # If n_conf is defined, read multiple lines as one configuration
    elif start is None and end is None and n_conf is not None:
        with open(filepath, "r") as f:
            i_line = 0
            conf = []
            for line in f:  # This actually reads line by line and only keeps the current line in memory
                # Ignore comments, separate by delimiter
                vals = split_line(line)
                line_values = []
                if vals:
                    for column in columns:
                        try:
                            value = vals[column]
                        except IndexError:
                            logger.warning("The given index '{}' could not be found on the line '{}'. The given delimiter or index could be wrong.".format(column, line))
                            return
                        try:
                            value = float(value)
                        except ValueError:
                            logger.warning("Could not cast value '{}' to float. Currently only floating point values are accepted".format(value))
                            return
                        else:
                            line_values.append(value)
                    conf.append(line_values)
                    i_line += 1
                    if i_line == n_conf:
                        yield np.array(conf)
                        conf = []
                        i_line = 0
